# -*- coding: utf-8 -*-

import cv2
import numpy as np
import glob
import os
import datetime
import re
import math
import time
import concurrent.futures
import shutil
import ctypes
import ContecCDIO
import serial
import methods
from distutils import dir_util
from operator import itemgetter
from configurations import *


# 倍率：20x-0.5x を前提として，捕捉位置を割り出す．数字を陽に設定していたり，かなり無理やりなのでいつか書き直す．
def where_to_track_auto():
    img_src = cv2.imread("./src/mfd.tif", cv2.IMREAD_COLOR)
    img_gry = cv2.cvtColor(img_src, cv2.COLOR_BGR2GRAY)
    circles = cv2.HoughCircles(img_gry, cv2.HOUGH_GRADIENT, 2, minDist=150, param1=100, param2=100, minRadius=70, maxRadius=80)

    for (x, y, r) in circles[0]:
        x, y, r = int(x), int(y), int(r)
        img_cir = cv2.circle(img_src, (x, y), r, (255, 255, 0), 1)
    cv2.imwrite("Hough.png", img_cir)

    # 巣と巣の間の距離を計算する
    centers_x = [cir[0] for cir in circles[0]]
    centers_y = [cir[1] for cir in circles[0]]
    centers_r = [cir[2] for cir in circles[0]]

    sorted_x = sorted(centers_x)
    sorted_y = sorted(centers_y)
    average_r = sum(centers_r) / len(centers_r)
    dif_x = [sorted_x[i + 1] - sorted_x[i] for i in range(len(sorted_x) - 1 ) if 2 * average_r < sorted_x[i + 1] - sorted_x[i] < 4 * average_r]
    dif_y = [sorted_y[i + 1] - sorted_y[i] for i in range(len(sorted_y) - 1 ) if 2 * average_r < sorted_y[i + 1] - sorted_y[i] < 4 * average_r]
    if len(dif_x) == 0 or len(dif_y) == 0:
        print("Too many fails on Hough transformation.  The source image might be unsuitable. Retake the image and try again.")
        exit(0)
    average_dif_x = int(sum(dif_x) / len(dif_x))
    average_dif_y = int(sum(dif_y) / len(dif_y))

    # 巣間の距離を正確に求める
    centers = [(cir[0], cir[1]) for cir in circles[0]]
    sorted_centers_for_y = sorted(centers, key=itemgetter(1))
    row =0
    column = 0
    temporal_x_list = []
    temporal_y_list = []
    for i in range(len(sorted_centers_for_y) - 1):
        drift_y = sorted_centers_for_y[i + 1][1] - sorted_centers_for_y[i][1]
        temporal_x_list.append(sorted_centers_for_y[i][0])
        temporal_y_list.append(sorted_centers_for_y[i][1])
        if drift_y > 0.5 * average_dif_y and column == NUM_COLUMNS - 1:
            dif_x = (max(temporal_x_list) - min(temporal_x_list)) / column
            tilt_sign = (int(temporal_y_list[-1] - temporal_y_list[0] > 0) - int(temporal_y_list[-1] - temporal_y_list[0] < 0)) * \
                        (int(temporal_x_list[-1] - temporal_x_list[0] > 0) - int(temporal_x_list[-1] - temporal_x_list[0] < 0))

            y_corr_for_dif_x = (max(temporal_y_list) - min(temporal_y_list)) / column * tilt_sign
            break
        elif drift_y > 0.5 * average_dif_y and column < NUM_COLUMNS - 1:
            row += 1
            column = 0
            temporal_x_list = []
            temporal_y_list = []
            continue
        elif row == NUM_LINES - 1 and column == NUM_COLUMNS - 2:
            dif_x = (max(temporal_x_list) - min(temporal_x_list)) / column
            tilt_sign = (int(temporal_y_list[-1] - temporal_y_list[0] > 0) - int(temporal_y_list[-1] - temporal_y_list[0] < 0)) * \
                        (int(temporal_x_list[-1] - temporal_x_list[0] > 0) - int(temporal_x_list[-1] - temporal_x_list[0] < 0))

            y_corr_for_dif_x = (max(temporal_y_list) - min(temporal_y_list)) / column * tilt_sign
        column += 1

    sorted_centers_for_x = sorted(centers, key=itemgetter(0))
    row =0
    column = 0
    temporal_x_list = []
    temporal_y_list = []
    for i in range(len(sorted_centers_for_x) - 1):
        drift_x = sorted_centers_for_x[i + 1][0] - sorted_centers_for_x[i][0]
        temporal_x_list.append(sorted_centers_for_x[i][0])
        temporal_y_list.append(sorted_centers_for_x[i][1])
        if drift_x > 0.5 * average_dif_x and column == NUM_LINES - 1:
            dif_y = (max(temporal_y_list) - min(temporal_y_list)) / column
            tilt_sign = (int(temporal_y_list[-1] - temporal_y_list[0] > 0) - int(temporal_y_list[-1] - temporal_y_list[0] < 0)) * \
                        (int(temporal_x_list[-1] - temporal_x_list[0] > 0) - int(temporal_x_list[-1] - temporal_x_list[0] < 0))
            x_corr_for_dif_y = (max(temporal_x_list) - min(temporal_x_list)) / column * tilt_sign
            break
        elif drift_x > 0.5 * average_dif_x and column < NUM_LINES - 1:
            row += 1
            column = 0
            temporal_x_list = []
            temporal_y_list = []
            continue
        elif row == NUM_COLUMNS - 1 and column == NUM_LINES - 2:
            dif_y = (max(temporal_y_list) - min(temporal_y_list)) / column
            tilt_sign = (int(temporal_y_list[-1] - temporal_y_list[0] > 0) - int(temporal_y_list[-1] - temporal_y_list[0] < 0)) * \
                        (int(temporal_x_list[-1] - temporal_x_list[0] > 0) - int(temporal_x_list[-1] - temporal_x_list[0] < 0))
            x_corr_for_dif_y = (max(temporal_x_list) - min(temporal_x_list)) / column * tilt_sign
        column += 1

    try:
        print(x_corr_for_dif_y, y_corr_for_dif_x)
    except UnboundLocalError:
        print("Too many fails on Hough transformation. The source image might be unsuitable. Retake the image and try again.")
        exit(0)

    for cir in circles[0]:
        x = int(cir[0]) + 55
        y = int(cir[1])
        if y > 300:
            cand_pts = [(x, y)]
            break
    count = 0
    x = cand_pts[0][0]
    y = cand_pts[0][1]
    while True:
        count += 1
        x1 = x - dif_x * count
        x2 = x + dif_x * count
        if x1 > 0 and x2 < img_src.shape[1]:
            cand_pts.append((int(x1), int(y - y_corr_for_dif_x * count)))
            cand_pts.append((int(x2), int(y + y_corr_for_dif_x * count)))
        elif x1 > 0 and x2 > img_src.shape[1]:
            cand_pts.append((int(x1), int(y - y_corr_for_dif_x * count)))
        elif x1 < 0 and x2 < img_src.shape[1]:
            cand_pts.append((int(x2), int(y + y_corr_for_dif_x * count)))
        else:
            break

    pts = list(set(cand_pts))
    for pt in pts:
        count = 0
        while True:
            count += 1
            y1 = pt[1] - dif_y * count
            y2 = pt[1] + dif_y * count
            if y1 > 300 and y2 < img_src.shape[0]:
                cand_pts.append((int(pt[0] + x_corr_for_dif_y * count), int(y1)))
                cand_pts.append((int(pt[0] - x_corr_for_dif_y * count), int(y2)))
            elif y1 > 300 and y2 > img_src.shape[0]:
                cand_pts.append((int(pt[0] + x_corr_for_dif_y * count), int(y1)))
            elif y1 < 300 and y2 < img_src.shape[0]:
                cand_pts.append((int(pt[0] - x_corr_for_dif_y * count), int(y2)))
            else:
                break

    pts = list(set(cand_pts))
    img_zero = np.zeros_like(img_gry)
    for pt in pts:
        img_roi = cv2. circle(img_zero, pt, 50, 255, -1)

    img_and = cv2.bitwise_and(img_roi, img_gry)
    cv2.imwrite("_region-check.png", img_and)
    img_cir = cv2.circle(img_src, cand_pts[0], 10, (255, 255, 0), 1)
    cv2.imwrite("Hough.png", img_cir)

    return pts


def search_image_during_trapping(number):
    while True:
        imgs_path = glob.glob("./a.tif")
        names, times = Get_img_info(imgs_path)
        if len(imgs_path) != 0:
            camera.regulate_state(READY)
            Leave_a_comment_in(LOG_FILE,
                               "\n" + str(time.time() - BEGINNING) + "\timage:" + str(number) + " was taken\n")
            break
        else:
            time.sleep(0.1)
            pass
    return names, times, imgs_path


def make_binary_during_trapping(imgs):
    kernels = [np.ones((3, 3), np.uint8), np.ones((5, 5), np.uint8)]

    imgs_bw = []
    for img in imgs:
        file_name = img[2:]
        img_src = cv2.imread(file_name, 1)
        img_split = cv2.split(img_src)
        img_bw = Otsu_threshold(img_split[MEM_COLOR])

        img_opn = cv2.morphologyEx(img_bw, cv2.MORPH_OPEN, kernels[0])
        img_cls = cv2.morphologyEx(img_opn, cv2.MORPH_CLOSE, kernels[1])

        imgs_bw.append(img_cls)

    return imgs_bw


def count_trapped(img, img_roi, number):
    img_and = cv2.bitwise_and(img, img, mask=img_roi)
    label_number, img_lbl, contours, center_masses = cv2.connectedComponentsWithStats(img_and)
    pts_gv = []
    sizes_gv = []
    img_test = img
    for i in range(len(contours)):
        if GV_LOWER_SIZE < contours[i][-1] < GV_UPPER_SIZE:
            pts_gv.append((int(center_masses[i][0]), int(center_masses[i][1])))
            sizes_gv.append(contours[i][-1])
            img_test = cv2.circle(img, pts_gv[-1], 30, 255, 2)
            x = pts_gv[-1][0]
            y = pts_gv[-1][1]
            cv2.putText(img, str(i), (x+10, y+10), cv2.FONT_HERSHEY_COMPLEX, 0.5, 255)
        else:
            pass
    file_name = "_circle_" + str("%08d"%number) + ".png"      # GVとして検出したオブジェクトの確認
    cv2.imwrite(file_name, img_test)
    imgs_path = glob.glob("./*circle*.png")
    time.sleep(0.1)
    Img_remover_edt(imgs_path, "./check-trap___")       # 出力画像は別の場所に移動

    return pts_gv, sizes_gv


def refresh_device():
    Leave_a_comment_in(LOG_FILE, "\n" + str(time.time() - BEGINNING) + "\t____________Start device refresh____________\n\n")
    camera.regulate_state(REFRESH_DEVICE)

    valve.change_state_to(VALVE_STATE["STABILIZE"])
    pump.change_flow_rate_to(valve, pump0_=3000, pump1_=3000)
    pump.stop(1)
    time.sleep(120)
    valve.change_state_to(VALVE_STATE["INDUCE"])
    pump.change_flow_rate_to(valve, pump0_=1000, pump1_=1000)
    pump.stop(1)
    time.sleep(60)
    pump.change_flow_rate_to(valve, pump0_=300, pump1_=300)
    pump.stop(1)
    time.sleep(60)
    pump.change_flow_rate_to(valve, pump0_=100, pump1_=100)
    pump.stop(1)
    time.sleep(60)

    camera.regulate_state(READY)
    Leave_a_comment_in(LOG_FILE, "\n" + str(time.time() - BEGINNING) + "\t____________Device refresh: Done____________\n\n")

    print("Device refresh: Done.")


def detect_vesicles(imgs_path, nests, number):
    os.mkdir("./check-trap___/" + str(number))
    kernels = [np.ones((3, 3), np.uint8), np.ones((5, 5), np.uint8)]

    increment = 0
    for img_path in imgs_path:
        img = cv2.imread(img_path[2:], cv2.IMREAD_COLOR)
        img_split = cv2.split(img)
        count = 0   # 巣の名前は機械的に番号で振る
        pts_gv = []
        sizes_gv = []
        cirs_gv = []
        ##################################################
        # 巣を中心にトリミングしてGVの有無をチェックする #
        ##################################################
        for nest in nests:
            ###################################
            # まず画像を二値化し， ブロブ解析 #
            ###################################
            img_R = img_split[MEM_COLOR]
            img_src = img_R[nest[1]-60:nest[1]+60, nest[0]-60:nest[0]+60]
            img_bw = Otsu_threshold(img_src)

            img_opn = cv2.morphologyEx(img_bw, cv2.MORPH_OPEN, kernels[0])
            img_cls = cv2.morphologyEx(img_opn, cv2.MORPH_CLOSE, kernels[1])
            cv2.imwrite(str(count) + ".png", img_cls)
            Img_remover_edt(str(count) + ".png", "./check-trap___/" + str(number))

            label_number, img_lbl, data, center_masses = cv2.connectedComponentsWithStats(img_cls)
            # 背景とGV以外のブロブがあった場合は，適切に二値化できない＝GVがいないと判断する
            if label_number != 2:
                count += 1
                continue

            ############################################
            # GVとして適切な大きさのものについてのみ， #
            # 周長とサイズをもとに，球形度を計算       #
            # 球形度が一定以上高いものをGVと認定する   #
            ############################################
            for i in range(len(data)):
                x, y, w, h, size = data[i]
                if not GV_LOWER_SIZE < size < GV_UPPER_SIZE:
                    continue

                img_cut = img_cls[y:y+h, x:x+w]
                img_cnt, contours, hierarchy = cv2.findContours(img_cut, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
                max_len = 0
                for contour in contours:
                    length = cv2.arcLength(contour, True)
                    if max_len < length:
                        max_len = length
                if max_len == 0:
                    count += 1
                    continue
                circularity = 4.0 * np.pi * size / (max_len * max_len)
                if circularity > 0.7:
                    cirs_gv.append(circularity)
                    pts_gv.append((int(center_masses[i][0] + nest[0] - 60), int(center_masses[i][1] + nest[1] - 60)))
                    sizes_gv.append(size)
                    x = pts_gv[-1][0]
                    y = pts_gv[-1][1]
                    cv2.putText(img, str(count), (x, y), cv2.FONT_HERSHEY_COMPLEX, 0.5, (255, 0, 0))
            count += 1

        cv2.imwrite("check_" + str("%08d" % (number + increment)) + ".png", img)
        imgs_path = glob.glob("./*check_*.png")
        time.sleep(0.1)
        Img_remover_edt(imgs_path, "./check-trap___")  # 出力画像は別の場所に移動
        increment += 1

    return pts_gv, sizes_gv, cirs_gv


# ベシクルの捕捉状況の推移を観察するためのメソッド
def vesicle_trap(slp_time, cycle):
    Leave_a_comment_in(LOG_FILE, "\n" + str(time.time() - BEGINNING) + "\t____________method: vesicle-trap START____________\n\n")

    valve.change_state_to(VALVE_STATE["INDUCE"])
    pump.change_flow_rate_to(valve, pump0_=TRAP_FLOW, pump1_=TRAP_FLOW)

    fname_trap = "trap-ratio_" + str(cycle) + ".csv"
    with open(fname_trap, "w") as of_trap:
        nests = where_to_track_auto()  # GVの捕捉位置を設定する

        count = 0
        error_count = 0
        number = 0
        trap = False

        while trap is False:
            time.sleep(slp_time)
            camera.regulate_state(TAKE_PICTURE)
            names, times, imgs_path = search_image_during_trapping(number)
            pts_gv, sizes_gv, cirs_gv = detect_vesicles(imgs_path, nests, number)
            trap_ratio = len(pts_gv) / NEST_NUMBER
            if 1 >= trap_ratio >= 0.25:
                count += 1
                error_count = 0
            else:
                error_count += 1
                count = 0

            print(str(count) + "\ttrap ratio is now: " + str(trap_ratio))
            of_trap.write(str(times[0]) + "," + str(trap_ratio) + "\n")

            # 捕捉率，保持時間ともに十分なら次に進む
            if count * slp_time >= TIME_OF_STABILIZED_TRAP:
                trap = True
            elif error_count > 60:  # 1時間1/4以下なら、つまりが発生したと判断して、一度デバイスをリフレッシュ
                Leave_a_comment_in(LOG_FILE, "\n" + str(time.time() - BEGINNING) + "\t____________Start device refresh____________\n\n")
                camera.regulate_state(REFRESH_DEVICE)
                refresh_device()
                camera.regulate_state(READY)
                Leave_a_comment_in(LOG_FILE, "\n" + str(time.time() - BEGINNING) + "\t____________Device refresh: Done____________\n\n")
                error_count = 0
                valve.change_state_to(VALVE_STATE["INDUCE"])
                pump.change_flow_rate_to(valve, pump0_=TRAP_FLOW, pump1_=TRAP_FLOW)

            time.sleep(0.1)
            Img_remover_org(imgs_path[0], number, "./org-trap___")  # 処理した画像は保存フォルダに移動

            number += 1
        # GVが十分な数に達したら，流れを切り，5分間安定化させる．
        valve.change_state_to(VALVE_STATE["STABILIZE"])
        pump.change_flow_rate_to(valve, pump0_=TRAP_FLOW, pump1_=3600)

        stat = time.time()
        while time.time() - stat < 300:
            time.sleep(slp_time)
            camera.regulate_state(TAKE_PICTURE)
            names, times, imgs_path = search_image_during_trapping(number)
            pts_gv, sizes_gv, cirs_gv = detect_vesicles(imgs_path, nests, number)
            trap_ratio = len(pts_gv) / NEST_NUMBER
            if 1 >= trap_ratio >= 0.25:
                count += 1
                error_count = 0
            else:
                error_count += 1
                count = 0

            print(str(count) + "\ttrap ratio is now: " + str(trap_ratio))
            of_trap.write(str(times[0]) + "," + str(trap_ratio) + "\n")
            time.sleep(0.1)
            Img_remover_org(imgs_path[0], number, "./org-trap___")  # 処理した画像は保存フォルダに移動

            number += 1
    Leave_a_comment_in(LOG_FILE, "\n" + str(time.time() - BEGINNING) + "\t____________method: vesicle-trap FINISH____________\n\n")
    pump.stop(1)
    time.sleep(300)

    return pts_gv, sizes_gv, times, number, cirs_gv


def make_dirs():
    if not os.path.exists("./org-trap___"):
        os.mkdir("./org-trap___")
    if not os.path.exists("./org-observation___"):
        os.mkdir("./org-observation___")
    if not os.path.exists("./check-trap___"):
        os.mkdir("./check-trap___")
    if not os.path.exists("./org-condensation___"):
        os.mkdir("./org-condensation___")


def record_results(vesicles_, cycle_):
    for ves in vesicles_:
        ves.record_data()

    dirs = glob.glob("./*___*")
    if not os.path.exists("./cycle-" + str(cycle_)):
        os.mkdir("./cycle-" + str(cycle_))
    for dir_ in dirs:
        try:
            shutil.move(dir_, "./cycle-" + str(cycle_))
        except shutil.Error:
            time.sleep(0.1)
            shutil.move(dir_, "./cycle-" + str(cycle_))
        except PermissionError:
            time.sleep(0.1)
            shutil.move(dir_, "./cycle-" + str(cycle_))


def shut_down_the_system():
    camera.regulate_state(STOP_MANSIONS)
    valve.change_state_to(VALVE_STATE["STABILIZE"])
    valve.exit()
    pump.stop(0)
    pump.stop(1)
    pump.exit()


def main(cycle):
    #############################################################################
    # By "Ctrl + C", the code is abruptly interpreted.                          #
    # e.g. refreshing the device during the measurement and restart the method. #
    #############################################################################
    hist = cycle
    try:
        make_dirs()  # make directory to reserve each captured images
        pts_gv, sizes_gv, times, number, cirs_gv = vesicle_trap(slp_time=60, cycle=cycle)

        vesicles = []
        imgs = sorted(glob.glob("./org-trap___/*.tif"))
        for i in range(len(pts_gv)):
            ves_temp = Vesicle(i, times[-1], pts_gv[i], sizes_gv[i], imgs[-1], cirs_gv[i])
            vesicles.append(ves_temp)

        ###############################################################
        # Write individual code to measure the dynamics of liposomes. #
        # The method code are written in "methods.py".                #
        ###############################################################
        number = 0
        # Currently, slp_time should be longer than 15 sec due to the processing time for the image analysis (typically 2-5 sec).
        if cycle < REPEATING * len(MIXING_RATIO) * len(EXCHANGE_RATE):
            # Simple timelapse during the exposure to uranine solution. Interval: 5 min, duration: 60 min.
            wash, number = methods.condensation_test_simple_timelapse(pump, valve, camera, vesicles, slp_time=300, number=number, cycle_=cycle)
        else:
            # Evaluate the concentration of uranine after 30 min of exposure to uranine solution.
            wash, number = methods.condensation_test_steadied_concentration(pump, valve, camera, vesicles, number=number, cycle_=cycle)
        print(str(number) + " images were taken.")

        record_results(vesicles, cycle)

        if wash is REQUIRED:
            # preliminary wash to avoid the back flow of substances into liposome channel.
            pump.change_flow_rate_to(valve, pump0_=40, pump1_=40)
            pump.stop(1)
            valve.change_state_to(VALVE_STATE["STABILIZE"])
            time.sleep(900)
        elif wash is NOT_REQUIRED:
            pass
        refresh_device()
        hist = cycle + 1

    except KeyboardInterrupt:
        code = "a"
        while code != "refresh" and code != "restart" and code != "exit":
            print("refresh: refresh and restart the cycle")
            print("restart: restart the cycle without refresh")
            print("exit: stop the program")
            code = input(">>>")

        if code == "refresh":
            refresh_device()
        elif code == "restart":
            pass
        elif code == "exit":
            Leave_a_comment_in(LOG_FILE, "Shut down the program by the code of Ctrl + C.")
            shut_down_the_system()
            exit(0)
        else:
            Leave_a_comment_in(LOG_FILE, "unexpected code was detected. shut down the program.")
            print("unexpected code was detected.")
            shut_down_the_system()
            exit(0)

        Leave_a_comment_in(LOG_FILE, "\n" + str(time.time() - BEGINNING) + "\t_________ReStart CYCLE " + str(_i_) + "_________\n\n")
        main(hist)
        Leave_a_comment_in(LOG_FILE, "\n" + str(time.time() - BEGINNING) + "\t_________ReFINISH CYCLE " + str(_i_) + "_________\n\n")


if __name__ == "__main__":
    #####################################################################################
    # This code is the main script.                                                     #
    # Individual method codes are separated in "methods.py".                            #
    # Configuration parameters and common functions are together in "configuration.py". #
    # The first letter is capitalized for functions in the configuration file           #
    # Header file for contec is "contecCDIO.py"                                         #
    #####################################################################################
    if os.path.exists("./log.txt") or os.path.exists("./flow_log.csv"):
        print("File already exists. Check again.")
        exit(0)

    Leave_a_comment_in(LOG_FILE, str(datetime.datetime.now()) + " program started")
    Leave_a_comment_in(LOG_FILE, "Elapsed time\tEvent\n")

    Leave_a_comment_in(LOG_FL_FILE, "elapsed time, flow\n")
    time.sleep(1)

    ##################
    # Initialization #
    ##################
    valve = Valve()
    valve.change_state_to(VALVE_STATE["STABILIZE"])
    time.sleep(0.5)
    pump = Pump()
    camera = Camera()
    time.sleep(0.1)
    print("\nInitialization Done.")
    Leave_a_comment_in(LOG_FILE, "\n" + str(time.time() - BEGINNING) + "\tSystem initialization: Done\n")

    ###############
    # Measurement #
    ###############
    for _i_ in range(NUMBER_OF_CYCLES):
        Leave_a_comment_in(LOG_FILE, "\n" + str(time.time() - BEGINNING) + "\t_________Start CYCLE " + str(_i_) + "_________\n\n")
        main(_i_)
        Leave_a_comment_in(LOG_FILE, "\n" + str(time.time() - BEGINNING) + "\t_________FINISH CYCLE " + str(_i_) + "_________\n\n")

    ########################
    # Shut down the system #
    ########################
    shut_down_the_system()
    print("\nThe end of the program.\n")
    Leave_a_comment_in(LOG_FILE, "\n\n\n" + str(time.time() - BEGINNING) + "\tEvery-event finished successfully.\n\n")
