# -*- coding: utf-8 -*-


import cv2
import numpy as np
import glob
import os
import datetime
import re
import math
import time
import concurrent.futures
import shutil
import ctypes
import ContecCDIO
import serial
from distutils import dir_util
from operator import itemgetter


class Pump:
    def __init__(self):
        self.pump0 = 0
        self.pump1 = 0
        self.ser = serial.Serial("COM5", baudrate=9600)
        self.ser.close()
        self.ser.open()
        self.ser.reset_input_buffer()
        if self.ser.out_waiting > 0:
            self.ser.reset_output_buffer()
        Leave_a_comment_in(LOG_FILE, str(time.time() - BEGINNING) + "\tPump initialization: Done\n")

        self.start(pump0_=TRAP_FLOW, pump1_=TRAP_FLOW)

    def start(self, pump0_, pump1_):
        self.pump0 = pump0_
        self.pump1 = pump1_
        comm0 = "0irate " + str(self.pump0) + " ul/hr\r"
        comm1 = "1irate " + str(self.pump1) + " ul/hr\r"
        self.ser.write(bytes(comm0, "UTF-8"))
        time.sleep(0.1)
        self.ser.write(bytes(comm1, "UTF-8"))
        time.sleep(0.1)

        comm0 = "0tvolume 5 ml\r"
        comm1 = "1tvolume 5 ml\r"
        self.ser.write(bytes(comm0, "UTF-8"))
        time.sleep(0.1)
        self.ser.write(bytes(comm1, "UTF-8"))
        time.sleep(0.1)

        comm0 = "0run\r"
        comm1 = "1run\r"
        self.ser.write(bytes(comm0, "UTF-8"))
        time.sleep(0.1)
        self.ser.write(bytes(comm1, "UTF-8"))

        print("\nPump started.")
        print("\nFlow condition was changed to")
        print("\tpump0: " + str(self.pump0) + "uL/hr")
        print("\tpump1: " + str(self.pump1) + "uL/hr")
        
        Leave_a_comment_in(LOG_FILE, str(time.time() - BEGINNING) + "\tPump started. pump0: " + str(self.pump0) + "pump1: " + str(self.pump1) + "\n")
        Leave_a_comment_in(LOG_FL_FILE, str(time.time() - BEGINNING) + "," + str(self.pump0) + "\n")

    def exit(self):
        self.ser.close()
        print("Pump exited.")
        Leave_a_comment_in(LOG_FILE, str(time.time() - BEGINNING) + "\tPump exited.\n")

    def change_flow_rate_to(self, valve, pump0_, pump1_):
        if valve.state == "MIXING" or "CONST-EXCHANGE":
            Leave_a_comment_in(LOG_FL_FILE, str(time.time() - BEGINNING) + "," + str(pumpC) + "\n")
        else:
            Leave_a_comment_in(LOG_FL_FILE, str(time.time() - BEGINNING) + "," + str(pump.pump0) + "\n")

        self.pump0 = pump0_
        self.pump1 = pump1_
        comm0 = "0stop\r"
        comm1 = "1stop\r"
        self.ser.write(bytes(comm0, "UTF-8"))
        time.sleep(0.1)
        self.ser.write(bytes(comm1, "UTF-8"))
        time.sleep(0.1)

        comm0 = "0irate " + str(self.pump0) + " ul/hr\r"
        comm1 = "1irate " + str(self.pump1) + " ul/hr\r"
        self.ser.write(bytes(comm0, "UTF-8"))
        time.sleep(0.1)
        self.ser.write(bytes(comm1, "UTF-8"))
        time.sleep(0.1)

        comm0 = "0tvolume 5 ml\r"
        comm1 = "1tvolume 5 ml\r"
        self.ser.write(bytes(comm0, "UTF-8"))
        time.sleep(0.1)
        self.ser.write(bytes(comm1, "UTF-8"))
        time.sleep(0.1)

        comm0 = "0run\r"
        comm1 = "1run\r"
        self.ser.write(bytes(comm0, "UTF-8"))
        time.sleep(0.1)
        self.ser.write(bytes(comm1, "UTF-8"))
        time.sleep(0.1)

        print("\nFlow condition was changed to")
        print("\tpump0: " + str(self.pump0) + "uL/hr")
        print("\tpump1: " + str(self.pump1) + "uL/hr")
        Leave_a_comment_in(LOG_FILE, str(time.time() - BEGINNING) + "\tPump flow rate was changed. pump0: " + str(self.pump0) + ", pump1: " + str(self.pump1) + "\n")

        if valve.state == "MIXING" or "CONST-EXCHANGE":
            Leave_a_comment_in(LOG_FL_FILE, str(time.time() - BEGINNING) + "," + str(pumpC) + "\n")
        else:
            Leave_a_comment_in(LOG_FL_FILE, str(time.time() - BEGINNING) + "," + str(pump.pump0) + "\n")

    def mixing(self, ratio):
        comm0 = "0stop\r"
        comm1 = "1stop\r"
        self.ser.write(bytes(comm0, "UTF-8"))
        time.sleep(0.1)
        self.ser.write(bytes(comm1, "UTF-8"))
        time.sleep(0.1)

        self.pump0 = 30000
        self.pump1 = 30000
        comm0 = "0irate " + str(self.pump0) + " ul/hr\r"
        comm1 = "1irate " + str(self.pump1) + " ul/hr\r"
        self.ser.write(bytes(comm0, "UTF-8"))
        time.sleep(0.1)
        self.ser.write(bytes(comm1, "UTF-8"))
        time.sleep(0.1)

        comm0 = "0tvolume " + str(500 * (1 - ratio)) + " ul\r"
        comm1 = "1tvolume " + str(500 * ratio) + " ul\r"
        self.ser.write(bytes(comm0, "UTF-8"))
        time.sleep(0.1)
        self.ser.write(bytes(comm1, "UTF-8"))
        time.sleep(0.1)

        comm0 = "0run\r"
        comm1 = "1run\r"
        self.ser.write(bytes(comm0, "UTF-8"))
        time.sleep(0.1)
        self.ser.write(bytes(comm1, "UTF-8"))
        time.sleep(0.1)
        time.sleep(60)

    def stop(self, address):
        comm = ["0stop\r", "1stop\r"]
        self.ser.write(bytes(comm[address], "UTF-8"))

        print("pump" + str(address) + " stopped.")
        Leave_a_comment_in(LOG_FILE, str(time.time() - BEGINNING) + "\tPump" + str(address) + " stopped\n")


class Valve:
    def __init__(self):
        self.name = ctypes.create_string_buffer(b"DIO000")
        self.id = ctypes.c_short()
        self.dll = ctypes.windll.LoadLibrary("cdio.dll")
        self.state = None

        ret = self.dll.DioInit(self.name, ctypes.byref(self.id))

        if ret != ContecCDIO.DIO_ERR_SUCCESS:
            print("\nFailed to initialize valve. Try again.")
            Leave_a_comment_in(LOG_FILE, str(time.time()) + "\tValve initialization failed.\n")
            exit(0)
        else:
            print("\nValve initialization finished.")
            Leave_a_comment_in(LOG_FILE, str(time.time()-BEGINNING) + "\tValve initialization finished.\n")

    def exit(self):
        self.dll.DioExit(self.id)
        print("\nValve exited.\n")
        Leave_a_comment_in(LOG_FILE, str(time.time() - BEGINNING) + "\tValve exited.\n")

    def change_state_to(self, state_):
        if self.state == "MIXING" or "CONST-EXCHANGE":
            Leave_a_comment_in(LOG_FL_FILE, str(time.time() - BEGINNING) + "," + str(pumpC) + "\n")
        elif self.state is None:
            pass
        else:
            Leave_a_comment_in(LOG_FL_FILE, str(time.time() - BEGINNING) + "," + str(pump.pump0) + "\n")

        for key, value in VALVE_STATE.items():
            if value == state_:
                self.state = key

        ret = self.dll.DioOutByte(self.id, 0, state_)
        if ret == ContecCDIO.DIO_ERR_SUCCESS:
            print("Valve state was changed to " + self.state + " mode.")
            Leave_a_comment_in(LOG_FILE, str(time.time()-BEGINNING) + "\tValve state was changed to " + self.state + " mode.\n")
        elif ret == ContecCDIO.DIO_ERR_SYS_RECOVERED_FROM_STANDBY:
            self.dll.DioResetDevice(self.id)
            print("Reset the valve. Valve state was changed to INDUCE mode.")
            Leave_a_comment_in(LOG_FILE, str(time.time()-BEGINNING) + "\tValve was reset. Valve state is now INDUCE mode.\n")
        else:
            print("Failed to change the valve state.")
            Leave_a_comment_in(LOG_FILE, str(time.time()-BEGINNING) + "\tFailed to change the valve state.\n")

        if self.state == "MIXING" or "CONST-EXCHANGE":
            Leave_a_comment_in(LOG_FL_FILE, str(time.time() - BEGINNING) + "," + str(pumpC) + "\n")
        else:
            Leave_a_comment_in(LOG_FL_FILE, str(time.time() - BEGINNING) + "," + str(pump.pump0) + "\n")


class Camera:
    def __init__(self):
        self.state = READY
        self.regulate_state(self.state)

    def regulate_state(self, signal):
        self.state = signal
        height = 50
        width = 50
        img = np.zeros((height, width, 3), np.uint8)
        for h in range(height):
            for w in range(width):
                img[h, w] = (self.state, self.state, self.state)
        cv2.imwrite("state.bmp", img)
        imgs_path = glob.glob("./*bmp")
        Img_copy(imgs_path, "C:/Users/ToyotaLab/Desktop/do-not-delete")
        for img in imgs_path:
            os.remove(img)


class Vesicle:
    def __init__(self, number, time_, center_mass, size, path_img, cir):
        name = "ves-" + str(number)
        self.params = {"time": [time_], "CM": [center_mass], "size": [size], "age": 0, "name": name, "state": ALIVE,
                       "blue": [], "green": [], "red": [], "bg": [], "cir": [cir]}
        b, g, r = self.get_average_bgr(path_img)
        self.params["blue"].append(b[0])
        self.params["green"].append(g[0])
        self.params["red"].append(r[0])
        bg = self.get_background(path_img)
        self.params["bg"].append(bg)

        if not os.path.exists("./" + str(self.params["name"]) + "___"):
            os.mkdir("./" + str(self.params["name"]) + "___")

        img_src = cv2.imread(path_img, 1)
        img_src_b, img_src_g, img_src_r = cv2.split(img_src)  # BGR分割し，RをもとにGVを検出

        x = self.params["CM"][-1][0]
        y = self.params["CM"][-1][1]
        img_cut = img_src_r[y - 30:y + 30, x - 30:x + 30]  # ad-hoc adaptive binarizationのための切り出し
        img_bw = Otsu_threshold(img_cut)

        file_name = "_" + str(self.params["name"]) + "-" + str(self.params["age"]) + ".png"
        cv2.imwrite(file_name, img_bw)
        time.sleep(0.1)
        Img_remover_edt("./" + file_name, "./" + str(self.params["name"]) + "___")

    def get_background(self, path_img):
        img_src = cv2.imread(path_img, 1)
        img_src_b, img_src_g, img_src_r = cv2.split(img_src)  # BGR分割
        x = self.params["CM"][-1][0]
        y = self.params["CM"][-1][1]
        img_cut = img_src_g[y - 10:y + 10, x - 50:x - 30]  # GVの横20＊20の領域をBGとする
        ave = np.mean(img_cut)

        return ave

    def append_data(self, t, center_mass, size, bs, gs, rs, bgs, cirs):
        if type(center_mass) == "list" and type(size) == "list":
            for i in range(len(center_mass)):
                self.params["time"].append(t[i])
                self.params["CM"].append(center_mass[i])
                self.params["size"].append(size[i])
                self.params["age"] += 1
                self.params["blue"].append(bs[i][0])
                self.params["green"].append(gs[i][0])
                self.params["red"].append(rs[i][0])
                self.params["bg"].append(bgs[i])
                self.params["cir"].append(cirs[i])
        elif type(center_mass) != "list" and type(size) != "list":
            self.params["time"].append(t)
            self.params["CM"].append(center_mass)
            self.params["size"].append(size)
            self.params["age"] += 1
            self.params["blue"].append(bs[0])
            self.params["green"].append(gs[0])
            self.params["red"].append(rs[0])
            self.params["bg"].append(bgs)
            self.params["cir"].append(cirs)
        else:
            print("types are different each other.check again.")    # for debug
            exit(1)

    def show_data(self):
        for i in range(self.params["age"]):
            print("center of mass:" + str(self.params["CM"][i]), "size:" + str(self.params["size"][i]))

    def record_data(self):
        with open(self.params["name"] + "___.csv", "w") as file:
            file.write("time,x,y,size,blue,green,red,bg,circularity\n")
            for i in range(self.params["age"] + 1):
                t = str(self.params["time"][i])
                x, y = self.params["CM"][i]
                size = str(self.params["size"][i])
                b = str(self.params["blue"][i])
                g = str(self.params["green"][i])
                r = str(self.params["red"][i])
                bg = str(self.params["bg"][i])
                cir = str(self.params["cir"][i])
                file.write(t + "," + str(x) + "," + str(y) + "," + size + "," + b + "," + g + "," + r + "," + bg + "," + cir + "\n")

    def get_average_bgr(self, path_img):
        # name_img, times = Get_img_info(path_img)
        img_src = cv2.imread(path_img, 1)

        x = self.params["CM"][-1][0]
        y = self.params["CM"][-1][1]
        img_cut = img_src[y - 30:y + 30, x - 30:x + 30]  # ad-hoc adaptive binarizationのための切り出し

        img_cut_split = cv2.split(img_cut)
        img_msk = Otsu_threshold(img_cut_split[MEM_COLOR])        # 膜のチャネルの情報からベシクルの位置を定義

        mean_r = cv2.mean(img_cut_split[RED], mask=img_msk)
        mean_g = cv2.mean(img_cut_split[GREEN], mask=img_msk)
        mean_b = cv2.mean(img_cut_split[BLUE], mask=img_msk)

        return mean_b, mean_g, mean_r

    def update_gv_info(self, paths_img):
        names_img, times = Get_img_info(paths_img)
        for i in range(len(names_img)):
            img_src = cv2.imread(names_img[i], 1)
            img_src_b, img_src_g, img_src_r = cv2.split(img_src)        # BGR分割し，RをもとにGVを検出
            x = self.params["CM"][-1][0]
            y = self.params["CM"][-1][1]
            img_cut = img_src_r[y-30:y+30, x-30:x+30]     # ad-hoc adaptive binarizationのための切り出し
            img_bw = Otsu_threshold(img_cut)

            file_name = "_" + str(self.params["name"]) + "-" + str(self.params["age"]) + ".png"
            cv2.imwrite(file_name, img_bw)
            time.sleep(0.1)
            Img_remover_edt("./" + file_name, "./" + str(self.params["name"]) + "___")

            pt = np.array([[x, y]])

            label_num, img_lbl, contours, center_masses = cv2.connectedComponentsWithStats(img_bw)

            if label_num == 2:      # ベシクルが1つ存在すればそれを記録する
                size = []
                cm = []
                dist = []
                for j in range(1, label_num):
                    size.append(contours[j][-1])
                    x_temp, y_temp = center_masses[j]
                    x_temp = x_temp + (x - 30)
                    y_temp = y_temp + (y - 30)
                    cm.append((int(x_temp), int(y_temp)))
                    cm_temp = np.array([[int(x_temp), int(y_temp)]])
                    dist.append(np.linalg.norm(cm_temp - pt))

                label = dist.index(min(dist))
                b, g, r = self.get_average_bgr(paths_img[i])
                bg = self.get_background(paths_img[i])

                img_cnt, contours, hierarchy = cv2.findContours(img_bw, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
                max_len = 0
                for contour in contours:
                    length = cv2.arcLength(contour, True)
                    if max_len < length:
                        max_len = length
                circularity = 4.0 * np.pi * size[label] / (max_len * max_len)

                self.append_data(times[i], cm[label], size[label], b, g, r, bg, circularity)
            elif label_num > 2:                # 将来的には、ここに、新たなベシクルがあるかどうかの判定を入れる.
                self.params["state"] = DEAD
            else:       # 背景のみ（label_num = 1）であれば，ベシクルがいないと判断し，状態をDEADに変更する）
                self.params["state"] = DEAD


def Leave_a_comment_in(file_name, comment):
    with open(file_name, "a") as log:
        log.write(comment)


def Img_remover_org(imgs, numbers, place):
    if type(imgs) == list:
        for i in range(len(imgs)):
            try:
                os.rename(imgs[i], "./" + str("%08d"%numbers[i]) + "__.tif")
            except OSError:
                time.sleep(0.1)
                os.rename(imgs[i], "./" + str("%08d"%numbers[i]) + "__.tif")
            except PermissionError:
                time.sleep(0.1)
                os.rename(imgs[i], "./" + str("%08d"%numbers[i]) + "__.tif")

    else:
        try:
            os.rename(imgs, "./" + str("%08d"%numbers) + "__.tif")
        except OSError:
            time.sleep(0.1)
            os.rename(imgs, "./" + str("%08d"%numbers) + "__.tif")
        except PermissionError:
            time.sleep(0.1)
            os.rename(imgs, "./" + str("%08d"%numbers) + "__.tif")

    imgs_path = glob.glob("./*__.tif")
    for img in imgs_path:
        try:
            shutil.copy(img, place)
            os.remove(img)
        except shutil.Error:
            time.sleep(0.1)
            shutil.copy(img, place)
            os.remove(img)
        except PermissionError:
            time.sleep(0.1)
            os.remove(img)


def Img_remover_edt(imgs, place):
    if type(imgs) == list:
        for img in imgs:
            try:
                shutil.copy(img, place)
                os.remove(img)
            except shutil.Error:
                time.sleep(0.1)
                shutil.copy(img, place)
                os.remove(img)
            except PermissionError:
                time.sleep(0.1)
                os.remove(img)
    else:
        try:
            shutil.copy(imgs, place)
            os.remove(imgs)
        except shutil.Error:
            time.sleep(0.1)
            shutil.copy(imgs, place)
            os.remove(imgs)
        except PermissionError:
            time.sleep(0.1)
            os.remove(imgs)


def Img_copy(imgs, place):
    for img in imgs:
        try:
            shutil.copy(img, place)
        except shutil.Error:
            time.sleep(0.1)
            shutil.copy(img, place)
        except PermissionError:
            time.sleep(0.1)
            shutil.copy(img, place)


def Calculate_background(img_src, vesicles):
    ave = 0
    img_src_b, img_src_g, img_src_r = cv2.split(img_src)  # BGR分割
    for ves in vesicles:
        x = ves.params["CM"][-1][0]
        y = ves.params["CM"][-1][1]
        img_cut = img_src_g[y - 10:y + 10, x - 50:x - 30]  # GVの横20＊20の領域をBGとする
        ave += np.mean(img_cut)
    ave /= len(vesicles)

    return ave


def Search_image_without_interval(camera, number):
    while True:
        imgs_path = glob.glob("./a.tif")
        if len(imgs_path) != 0:
            camera.regulate_state(READY)
            Leave_a_comment_in(LOG_FILE,
                               "\n" + str(time.time() - BEGINNING) + "\timage:" + str(number) + " was taken\n")
            number += 1
            break
        else:
            time.sleep(0.1)
            pass
    return imgs_path, number


def Search_image_with_interval(camera, number, slp_time):
    inner_timer = time.time()
    while True:
        now = time.time()
        if now - inner_timer > slp_time:
            if now - inner_timer - slp_time > 0.5:
                Leave_a_comment_in(LOG_FILE, str(cycle_) + "-" + str(number) + ": capture delay\n")
                print(str(cycle_) + "-" + str(number) + ": capture delay\n")
            camera.regulate_state(TAKE_PICTURE)
            break
        time.sleep(0.1)

    while True:
        imgs_path = glob.glob("./a.tif")
        if len(imgs_path) != 0:
            camera.regulate_state(READY)
            Leave_a_comment_in(LOG_FILE,
                               "\n" + str(time.time() - BEGINNING) + "\timage:" + str(number) + " was taken\n")
            number += 1
            break
        else:
            time.sleep(0.1)
            pass
    return imgs_path, number


def Otsu_threshold(img_src):
    ret, img_bw = cv2.threshold(img_src, 0, 255, cv2.THRESH_BINARY+cv2.THRESH_OTSU)

    return img_bw


def Get_img_info(imgs):
    name = []
    elps = []
    if type(imgs) == list:
        for img in imgs:
            name.append(img[2:])  # pathから "./" を抜いたものを名前と定義
            elps.append(os.path.getmtime(img) - BEGINNING)
    else:
        name.append(imgs[2:])  # pathから "./" を抜いたものを名前と定義
        elps.append(os.path.getmtime(imgs) - BEGINNING)

    return name, elps


BEGINNING = time.time()
NUM_LINES = 5
NUM_COLUMNS = 7
NEST_NUMBER = 28
GV_UPPER_SIZE = 500
GV_LOWER_SIZE = 50
ALIVE = True
DEAD = False
BLUE = 0
GREEN = 1
RED = 2
MEM_COLOR = RED
VALVE_STATE = {"INDUCE": 0x00, "STABILIZE": 0x04, "MIXING": 0x06, "EXCHANGE": 0x05, "FREE-PRESSURE": 0x07}
READY = 10
TAKE_PICTURE = 100
REFRESH_DEVICE = 200
STOP_MANSIONS = 255
TIME_OF_STABILIZED_TRAP = 300        # sec
EXIT_TIME = 30      # sec
SEARCH_INTERVAL = 1     # sec
REQUIRED = True
NOT_REQUIRED = False
LOG_FILE = "log.txt"
LOG_FL_FILE = "log_flow.csv"

TRAP_FLOW = 20
MIXING_RATIO = [0.3]  # pump0: 外水相，pump1: 混合するもの
EXCHANGE_RATE = [20]  # uL/h
pumpC = 3  # uL/h
REPEATING = 3
NUMBER_OF_CYCLES = REPEATING * len(MIXING_RATIO) * len(EXCHANGE_RATE) * 2


