# -*- coding: utf-8 -*-

import cv2
import numpy as np
import glob
import os
import datetime
import re
import math
import time
import concurrent.futures
import shutil
import ctypes
import ContecCDIO
import serial
from distutils import dir_util
from operator import itemgetter
from configurations import *


def condensation_test_low_concentration(pump, valve, camera, vesicles, slp_time, number, cycle_):
    wash = REQUIRED
    with open("background___.csv", "a") as bg_log:
        pump.change_flow_rate_to(valve, pump0_=EXCHANGE_RATE[int(cycle_ / len(MIXING_RATIO)) % (len(EXCHANGE_RATE))],
                                 pump1_=EXCHANGE_RATE[int(cycle_ / len(MIXING_RATIO)) % (len(EXCHANGE_RATE))])
        pump.stop(1)
        time.sleep(30)
        valve.change_state_to(VALVE_STATE["EXCHANGE"])

        stat = time.time()
        while time.time() - stat < 3600:  # 60分観察する
            imgs_path, number = Search_image_with_interval(camera, number, slp_time)

            for ves in vesicles:
                if ves.params["state"] == ALIVE:
                    ves.update_gv_info(imgs_path)
                else:
                    pass

            img = cv2.imread(imgs_path[-1])
            ave = Calculate_background(img, vesicles)
            for ves in vesicles:
                if ves.params["state"] == ALIVE:
                    bg_log.write(str(ves.params["time"][-1]) + "," + str(ave) + "\n")
                    break

            time.sleep(0.1)
            Img_remover_org(imgs_path[-1], number, "./org-condensation___")  # 処理した画像は保管場所へ移動
        Leave_a_comment_in(LOG_FILE, "\n" + str(time.time() - BEGINNING) + "\tObservation: done\n")

    return wash, number


def condensation_test_simple_timelapse(pump, valve, camera, vesicles, slp_time, number, cycle_):
    wash = REQUIRED
    with open("background___.csv", "a") as bg_log:
        pump.change_flow_rate_to(valve, pump0_=EXCHANGE_RATE[int(cycle_ / len(MIXING_RATIO)) % (len(EXCHANGE_RATE))],
                                 pump1_=EXCHANGE_RATE[int(cycle_ / len(MIXING_RATIO)) % (len(EXCHANGE_RATE))])
        pump.stop(1)
        time.sleep(30)

        valve.change_state_to(VALVE_STATE["EXCHANGE"])
        stat = time.time()
        while time.time() - stat < 1800:  # 30分観察する
            imgs_path, number = Search_image_with_interval(camera, number, slp_time)

            for ves in vesicles:
                if ves.params["state"] == ALIVE:
                    ves.update_gv_info(imgs_path)
                else:
                    pass

            img = cv2.imread(imgs_path[-1])
            ave = Calculate_background(img, vesicles)
            for ves in vesicles:
                if ves.params["state"] == ALIVE:
                    bg_log.write(str(ves.params["time"][-1]) + "," + str(ave) + "\n")
                    break

            time.sleep(0.1)
            Img_remover_org(imgs_path[-1], number, "./org-condensation___")  # 処理した画像は保管場所へ移動

        valve.change_state_to(VALVE_STATE["STABILIZE"])
        stat = time.time()
        while time.time() - stat < 1800:  # 30分観察する
            imgs_path, number = Search_image_with_interval(camera, number, slp_time)

            for ves in vesicles:
                if ves.params["state"] == ALIVE:
                    ves.update_gv_info(imgs_path)
                else:
                    pass

            img = cv2.imread(imgs_path[-1])
            ave = Calculate_background(img, vesicles)
            for ves in vesicles:
                if ves.params["state"] == ALIVE:
                    bg_log.write(str(ves.params["time"][-1]) + "," + str(ave) + "\n")
                    break

            time.sleep(0.1)
            Img_remover_org(imgs_path[-1], number, "./org-condensation___")  # 処理した画像は保管場所へ移動
        Leave_a_comment_in(LOG_FILE, "\n" + str(time.time() - BEGINNING) + "\tObservation: done\n")

        # STABILIZE modeでデバイスを低流量でフルクトースに置換。いきなり高流量で洗うと、リポソーム側にウラニンがいく
        pump.change_flow_rate_to(valve, pump0_=EXCHANGE_RATE[int(cycle_ / len(MIXING_RATIO)) % (len(EXCHANGE_RATE))],
                                 pump1_=EXCHANGE_RATE[int(cycle_ / len(MIXING_RATIO)) % (len(EXCHANGE_RATE))])
        pump.stop(1)
        valve.change_state_to(VALVE_STATE["STABILIZE"])
        time.sleep(600)
    return wash, number


def condensation_test_steadied_concentration(pump, valve, camera, vesicles, number, cycle_):
    wash = REQUIRED
    with open("background___.csv", "a") as bg_log:
        pump.change_flow_rate_to(valve, pump0_=EXCHANGE_RATE[int(cycle_ / len(MIXING_RATIO)) % (len(EXCHANGE_RATE))],
                                 pump1_=EXCHANGE_RATE[int(cycle_ / len(MIXING_RATIO)) % (len(EXCHANGE_RATE))])
        pump.stop(1)
        time.sleep(30)
        valve.change_state_to(VALVE_STATE["EXCHANGE"])
        time.sleep(1800)  # 30分外溶液を置換し，その後1枚画像を撮る．

        camera.regulate_state(TAKE_PICTURE)
        imgs_path, number = Search_image_without_interval(camera, number)

        for ves in vesicles:
            if ves.params["state"] == ALIVE:
                ves.update_gv_info(imgs_path)
            else:
                pass

        img = cv2.imread(imgs_path[-1])
        ave = Calculate_background(img, vesicles)
        for ves in vesicles:
            if ves.params["state"] == ALIVE:
                bg_log.write(str(ves.params["time"][-1]) + "," + str(ave) + "\n")
                break

        time.sleep(0.1)
        Img_remover_org(imgs_path[-1], number, "./org-condensation___")  # 処理した画像は保管場所へ移動
        Leave_a_comment_in(LOG_FILE, "\n" + str(time.time() - BEGINNING) + "\tObservation: done\n")

        # STABILIZE modeでデバイスを低流量でフルクトースに置換。いきなり高流量で洗うと、リポソーム側にウラニンがいく
        pump.change_flow_rate_to(valve, pump0_=EXCHANGE_RATE[int(cycle_ / len(MIXING_RATIO)) % (len(EXCHANGE_RATE))],
                                 pump1_=EXCHANGE_RATE[int(cycle_ / len(MIXING_RATIO)) % (len(EXCHANGE_RATE))])
        pump.stop(1)
        valve.change_state_to(VALVE_STATE["STABILIZE"])
        time.sleep(600)
    return wash, number


# 総観察時間とインターバルを指定し，注目するベシクルの候補を変数として，単純に時間発展を見る．
def simple_observation(pump, valve, camera, vesicles, tot_time, slp_time, number):
    wash = REQUIRED
    stat = time.time()
    while time.time()-stat < tot_time:
        time.sleep(slp_time)
        camera.regulate_state(TAKE_PICTURE)
        imgs_path, number = Search_image_without_interval(camera, number)
        for ves in vesicles:
            if ves.params["state"] == ALIVE:
                ves.update_gv_info(imgs_path)        # 当該関数内で処理した画像は保管場所へ移動
            else:
                pass
        time.sleep(0.1)
        Img_remover_org(imgs_path[-1], number, "./org-observation___")      # 処理した画像は保管場所へ移動
    return wash, number
